#!/usr/bin/env python3

"""
NAME:
	updrev.py - update rev maps

SYNOPSIS:
	updrev.py [options] ["named_data"]

DESCRIPTION:
	This is a tool to update reverse DNS zones.
	The "named_data" is output from getdata.py(1).
	If we are given 'named_dump.db' we run 'getdata.py' to
	get the format we want.
	We use the 'SOA', 'PTR' and 'A' records.

	'SOA' records tell us whether we are authoritative for the
	zone.

	'PTR' records tell us whether a mapping already exists in which
	case we leave it alone - provided that an A record for the
	same fqdn also exists. This allows an initial set of maps to
	be generated by this tool, then hand edited to reflect
	preferences which will then be left alone. 

	'A' records are used to generate new 'PTR' records on a first come
	basis. We first determine the IN-ADDR zone that the address
	belongs to and if we are authoritative for that zone, or the
	domain that the host is in, we generate a 'PTR' record.

	When checking for authority, we work our way up through the
	DNS hierarchy until we find an SOA record, or we hit "TOP_ZONE" or
	"TOP_ARPA_ZONE", if no SOA record is found the zone is
	considered unclaimed and if 'UNCLAIMED_REVZONE' is set to
	'take' we will pretend we own it.  This is only a good idea on
	a private network.
	
	Subsequent 'A' records for the same address, generate 'PTR'
	records which are commented out.  This makes it easier for a
	human hostmaster to pick a preferred PTR record from the
	available choices.

	If a file with the same name as a rev zone exists with a
	'.inc' extention (eg. rev/203.12.250.inc) it will be
	$INCLUEDed into the 'rev' zone.  This allows for non-PTR
	records for delegations and such to be kept apart from the
	automagically maintained '.db' and '.soa' files.

	For each new IN-ADDR zone we add an appropriate entry to
	"primary_zones".

	The new reverse maps are only installed if they differ from
	the previous version.  For this, we rely on the SOA records
	being kept in a separate file to the zone data.	 The SOA
	records are maintained by make(1) and updsoa.py(8).

	Options:

	-a "added"
		List added zone files in "added".  This can be handy
		to ensure they are added to an SCM (git, mercurial etc).
	
	-c "config"
		Read settings from "config", see conf.py for syntax.
		We will automatically read 'dns.cf' and '${progname}.cf'
		from the directory this script is in ('${progdir}'),
		as well as '.dnc.cf' and '.${progname}.cf' if they
		exist in the current directory or a parent.

		In fact unlike earlier iterations of this tool, we
		expect most settings to be provided via a "config" file.

	-d "dbg"
		Enable debug output for the specified operations.

	-N "primary_zones"
		This option is for backwards compatability only.
	
	-o "var=value"
		This allows overriding random settings from "config".
	
AUTHOR:
	Simon J. Gerraty <sjg@crufty.net>

"""

"""
RCSid:
	$Id: updrev.py,v 1.19 2023/06/06 05:42:03 sjg Exp $

	@(#) Copyright (c) 1996-2023 Simon J. Gerraty.

	This file is provided in the hope that it will
	be of use.  There is absolutely NO WARRANTY.
	Permission to copy, redistribute or otherwise
	use this file is hereby granted provided that 
	the above copyright notice and this notice are
	left intact. 
      
	Please send copies of changes and bug-fixes to:
	sjg@crufty.net

"""

import sys
import subprocess
import shutil
import ipaddress
import os
import os.path
import tempfile
import iputils
import dnsutils
import conf as cf

SOAs = {}                               # which zones do we control
OldPTRs = {}                            # pre-existing PTR records
PTRs = {}                               # candidates from A records
As = {}                                 # contains ';' if seen before

def check_soa(domain, conf={}, top=None):
    """indicate whether we are authoritative for domain
    return will be None if no entry found, True or False
    """
    global SOAs

    if domain in SOAs:
        s = SOAs[domain]
        dnsutils.debug_print('soa', 'check_soa: {}: {}'.format(domain, s), conf)
        return s
    parent = '.'.join(domain.split('.')[1:])
    if not top:
        if 'in-addr.arpa' in parent:
            top = conf.get('TOP_ARPA_ZONE', 'in-addr.arpa')
        elif 'ip6.arpa' in parent:
            top = conf.get('TOP_ARPA6_ZONE', 'ip6.arpa')
        else:
            top = conf.get('TOP_ZONE', '.')
        dnsutils.debug_print('soa', 'check_soa: {}: top={}'.format(domain,top), conf)
    if parent == top or '.' not in parent:
        dnsutils.debug_print('soa', 'check_soa: {}: not found'.format(domain), conf)
        return None
    return check_soa(parent, conf, top)

def skip_PTR(ip, fqdn, conf):
    skip_PTRs = conf.get('skip_PTRs', [])
    if ip in skip_PTRs:
        return True
    if fqdn in skip_PTRs:
        return True
    for p in skip_PTRs:
        if ip.endswith(p):
            return True
    return False

def skip_A(ip, fqdn, conf):
    skip_As = conf.get('skip_As', [])
    if ip in skip_As:
        return True
    if fqdn in skip_As:
        return True
    for p in skip_As:
        if ip.startswith(p):
            return True
    return False

def skip_zone(prefix, rzone, conf):
    skip_zones = conf.get('skip_zones', [])
    if prefix in skip_zones:
        return True
    if rzone in skip_zones:
        return True
    return False

def gen_ptr(ip,fqdn,conf={}):
    """generate candidate PTRs"""

    cidr = iputils.get_cidr_net(ip, conf)

    if cidr.is_ip6:
        ip6 = ipaddress.IPv6Address(ip)
        a = ip6.reverse_pointer.split('.')
        host = '.'.join(a[:cidr.host_nibbles])
        net = '.'.join(a[cidr.host_nibbles:])
        dnsutils.debug_print('ptrs', 'gen_ptr: cidr={} net={} host={} fqdn={}'.format(cidr.prefix, net, host, fqdn), conf)
        return (net,host,fqdn)
    
    prefixes = conf.get('cidr_prefixes', {})
    host_octets = conf.get('host_octets', {})
    if cidr.prefix in prefixes:
        net = prefixes[cidr.prefix]
        h = host_octets[cidr.prefix]
    else:
        net = cidr.prefix
        # how many octets for host?
        h = cidr.host_octets
        # if it is 8bit boundary, toss any trailing .0's
        if cidr.bits % 8 == 0:
            a = net.split('.')
            i = len(a) - 1
            while i > 0:
                if a[i] != '0':
                    break
                i -= 1
            net = '.'.join(a[:i+1])
            h -= 1
        host_octets[cidr.prefix] = h
        net = iputils.ip2arpa(net)
        prefixes[cidr.prefix] = net
    a = ip.split('.')
    a.reverse()
    host = '.'.join(a[:h+1])
    dnsutils.debug_print('ptrs', 'gen_ptr: cidr={} net={} host={} fqdn={}'.format(cidr.prefix, net, host, fqdn), conf)
    return (net,host,fqdn)

def get_ptrs(dfile, conf={}):
    """Get established and candidate PTRs from data file"""
    global SOAs
    global OldPTRs
    global PTRs
    global As

    ns_list = conf.get('ns_list')
    # networks we claim authority for
    nets = conf.get('nets', [])
    for net in nets:
        # this will populate cidr_nets used by get_mask
        c = iputils.CIDR(net, conf=conf)
        SOAs[c.in_addr_zone()] = True

    fd,conf['ptr_tmp'] = tempfile.mkstemp(text=True)
    ptout = os.fdopen(fd, 'w')
    for line in open(dfile, 'r').readlines():
        w = line.strip().split()

        if not w[0] in ['A','AAAA','PTR','SOA']:
            continue
        if w[0] == 'SOA':
            # SOA key ns email
            if w[2] in ns_list:
                SOAs[w[1]] = True
            else:
                SOAs[w[1]] = False
            continue
        host,domain = w[1].split('.', 1)
        if host == 'localhost':
            continue
        if w[1] in ['127.0.0.1', '1.0.0.127.in-addr.arpa']:
            continue
        if w[0] == 'PTR':
            if not check_soa(domain, conf):
                continue

            if 'ip6.arpa' in w[1]:
                ip = iputils.arpa2ip6(w[1])
            else:
                a = w[1].strip('.in-addr.arpa').split('.')
                if len(a) > 4:
                    # CIDR PTR ?
                    # Per rfc2317 we migh see
                    # n4.net/bits.n3.n2.n1
                    # or
                    # n4.net-bits.n3.n2.n1
                    # we want to skip the octet that is not digits
                    # regardless of where it appears.
                    n = []
                    for x in a:
                        if not x.isdigit():
                            continue
                        n.append(x)
                    a = n
                a.reverse()
                ip  = '.'.join(a)
            if skip_PTR(ip, w[2], conf):
                continue
            OldPTRs[ip] = gen_ptr(ip, w[2], conf)
            continue

        if w[0] in ['A', 'AAAA']:
            if w[0] == 'AAAA':
                # canonicalize it
                ip = str(ipaddress.IPv6Address(w[2]))
            else:
                ip = w[2]
            if ip in OldPTRs:
                (net,host,fqdn) = OldPTRs[ip]
                if fqdn == w[1]:
                    # we have confirmation of existing PTR
                    dnsutils.debug_print('ptrs', 'get_ptrs: confirmed: {} {} {}'.format(net,host,fqdn), conf)
                    print('{} {} {}'.format(net,host,fqdn), file=ptout)
                    PTRs[ip] = ''       # clobber other candidates
                    As[ip] = ';'        # we have done this one
                    continue

            if skip_A(ip, w[1], conf):
                continue
            c = As.get(ip, '')          # first time?
            As[ip] = ';'                # for next time
            rev = iputils.ip2arpa(ip)   # this will handle ip6 too
            authoritative = check_soa(rev, conf)
            if authoritative is None:
                # no one is authoritative but what about the domain?
                authoritative = check_soa(domain, conf)
                if authoritative is None:
                    if conf.get('UNCLAIMED_REVZONE', 'no') != 'take':
                        continue
            if authoritative is False:
                continue
            (net,host,fqdn) = gen_ptr(ip, w[1], conf)
            dnsutils.debug_print('ptrs', 'get_ptrs: {} {}{} {}'.format(net,c,host,fqdn), conf)
            if not c:
                # just remember for now
                PTRs[ip] = (net,host,fqdn)
                continue
            print('{} {}{} {}'.format(net,c,host,fqdn), file=ptout)

    for ip,ptr in PTRs.items():
        if not ptr:
            continue
        (net,host,fqdn) = ptr
        dnsutils.debug_print('ptrs', 'get_ptrs: {} {} {}'.format(net,host,fqdn), conf)
        print('{} {} {}'.format(net,host,fqdn), file=ptout)

    ptout.close()

def update_revs(conf={}):
    """Take the candidate PTRs and update zone files"""

    ptrtmp = conf.get('ptr_tmp')
    r = subprocess.Popen('sort -u {} | sort -n +1 -2'.format(ptrtmp),
                         shell=True,
                         universal_newlines=True,
                         stdout=subprocess.PIPE)
    added = conf.get('added')
    pzfile = conf.get('primary_zones', 'primary.zones')
    pzfile_new = pzfile+'.new'
    try:
        os.unlink(pzfile_new)
    except:
        pass
    if os.path.exists(pzfile):
        shutil.copy(pzfile, pzfile_new)
    pzout = open(pzfile_new, 'a')
    Opened = {}
    Zfiles = {pzfile_new: 1}
    zf = None
    lnet = None
    
    for line in r.stdout.readlines():
        (net,host,fqdn) = line.strip().split()

        if net != lnet:
            prefix = iputils.arpa2ip(net, False)
            rzone = '{}/{}'.format(conf.get('revdir', 'rev'), prefix.strip('::'))
            soa = rzone+'.soa'
            if skip_zone(prefix, rzone, conf):
                dnsutils.debug_print('revs', 'update_revs: skipping {} {}'.format(prefix, rzone), conf)
                continue
            rzfile = rzone+conf.get('ptr_zone_ext','.db')
            dnsutils.debug_print('revs', 'update_revs: rzone={}'.format(rzone), conf)
            if not net in Opened:
                if not os.path.exists(soa):
                    dnsutils.debug_print('revs', 'update_revs: soa={}'.format(soa), conf)
                    # for CIDR nets we need to ask it for in_addr_zone
                    c = iputils.get_cidr_net(net, conf)
                    print('\nzone "{}" {{\n\ttype primary;\n\tfile "{}";\n}};\n'.format(c.in_addr_zone(), soa), file=pzout)
                    if added:
                        print(rzfile, file=added)
                rzinc = rzone+'.inc'
                rzfile_new = rzfile+'.new'
                Opened[net] = zf = open(rzfile_new, 'w')
                Zfiles[rzfile_new] = 1
                dnsutils.debug_print('revs', 'update_revs: new: {}'.format(rzfile), conf)
                print("""; WARNING: This file is automagically maintained.
; Place all non-PTR records in {} and $INCLUDE it
; or they will dissappear.
;
""".format(os.path.basename(rzinc)), file=zf)
                if os.path.exists(rzinc):
                    print('$INCLUDE {};\n'.format(rzinc), file=zf)
            else:
                zf = Opened[net]
        lnet = net
        print('{}\tIN\tPTR\t{}.'.format(host, fqdn), file=zf)
    if zf:
        zf.close()
    pzout.close()
    if cf.getBool(conf, 'keep_tmp', False):
        dnsutils.debug_print('tmp', 'ptrs={}'.format(ptrtmp), conf)
    else:
        os.unlink(ptrtmp)

    dnsutils.install_new(Zfiles.keys(), conf)


def myopts(o,a,conf):
    if o == '-a':
        conf['added'] = open(a, 'a')
    elif o == '-N':                     # for backwards compat
        conf['primary_zones'] = a
        
if __name__ == '__main__':
    conf = {
        'DEBUG': {},
        'ns_list': [],
        }
    conf

    conf,args = dnsutils.app_init('a:N:', myopts, conf)

    if args[0].endswith('dump.db'):
        fd,data = tempfile.mkstemp(text=True)
        dout = os.fdopen(fd, 'w')
        r = subprocess.run([cf.subst('${progdir}/getdata.py', conf), args[0]],
                           universal_newlines=True,
                           stdout=dout)
        dout.close()
    else:
        fd = None
        data = args[0]

    get_ptrs(data, conf)
    update_revs(conf)
    if fd:
        os.unlink(data)
    sys.exit(0)
